//______________________________________________________________________________
//
//  Thread object with message support.
//  Not meant to use itself, but to derive other objects from.
//  Written for an article in Blaise Pascal Magazine
//
//  Feel free to alter and experiment with it.
//  Also feel free to ask questions or contribute or comment to this ideas.
//
//  J.P.Hoefnagel, 1M2M B.V.
//  EMail: jph@1m2m.eu
//______________________________________________________________________________
//

unit uBPM_BaseThread;

interface

uses
  System.Classes,System.SysUtils,system.UITypes,Graphics,
  winapi.windows,
  uBPM_Messages;

const SecPerDay = (24*60*60);

type
  TBPM_BaseThread = class(TThread)
  private
    fThrdTime   : int64;
    fSysTime    : int64;
    fSectmr     : tDateTime;
    fActiveReq  : boolean;
    fActiveAck  : boolean;
    fMsgHandle  : THandle;
    procedure SetMsgID(const Value: TBPM_MessageID);
  protected
    procedure   Execute; override;
  public
    constructor Create();
    function    OKToRun: boolean;
    procedure   OnMaint(var Done: Boolean); virtual;
    procedure   OnMaint1000();     virtual;
    procedure   OnActivate();      virtual;
    procedure   OnBeforeExecute(); virtual;
    procedure   OnAfterExecute();  virtual;
    procedure   OnDeActivate();    virtual;
    procedure   OnMsgHandler(mr:TBPM_MessageRec); virtual;
    procedure   OnErrorlog(Cde:integer;Msg:string;Color:TColor); virtual;
    property    Active:boolean read fActiveAck write fActiveReq;
    property    MsgID:TBPM_MessageID write SetMsgID;
    function    GetCPULoad: double;
  end;

var EndAllThreadsRequest:boolean;

implementation

{ TBPM_BaseThread }

constructor TBPM_BaseThread.create();
begin
  fActiveReq := false;
  fActiveAck := false;
  fMsgHandle := 0;    //  0 disables message handling from basethread
  fSectmr    := now();
  FreeOnTerminate:=true;
  inherited create();
end;

procedure TBPM_BaseThread.SetMsgID(const Value: TBPM_MessageID);
begin
  if fMsgHandle<>0 then begin           // if handle was open
    BPM_MsgBufClose(fMsgHandle);        //   then close it
    fMsgHandle:=0;                      //
  end;                                  //
  if Value>=-1 then                     // if new MsgID valid (-1..maxint)
    fMsgHandle := BPM_MsgBufOpen(Value);// then start listening to it
end;

function TBPM_BaseThread.OKToRun():boolean;
begin
  result:= (Terminated = false) and (EndAllThreadsRequest = false);
end;

procedure TBPM_BaseThread.Execute;
var mr:TBPM_MessageRec; Done:boolean;
begin
  OnBeforeExecute();
  while OKToRun() do try
    if (fActiveReq <> fActiveAck) then begin
      fActiveAck:=fActiveReq;
      if fActiveAck then OnActivate() else OnDeActivate();
    end;
    if fActiveAck then begin
      Done:=true; OnMaint(Done); // can set done to false to prevent sleep(20)
      if (Now()-fSecTmr)>=1/secperday then begin
        fSecTmr:=now(); OnMaint1000();
        Done:=false;
      end;
      if (fMsgHandle <> 0) then begin
        while BPM_MsgBufRead(fMsgHandle,MR) do begin
          OnMsgHandler(MR);
          Done:=false;
        end;
      end;
      if done then sleep(20); //reduces CPU load when thread is not too busy
    end else sleep(20);       //reduces CPU load when thread is not active
  except
    on e:exception do begin
      OnErrorlog(-1,'Error in '+self.ClassName+': '+e.message,clRed);
    end;
  end;
  OnAfterExecute();
end;

// fuction can be called from VCL and returns the CPU load cauzed by
// this threadobject in 0 to 100 %
function TBPM_BaseThread.GetCPULoad: double;{VCL}
var OThrdTime:int64;    OsysTime :int64;    T1,T2    :double;
var createtime,exittime,kerneltime,usertime:filetime;
begin
  if GetThreadTimes(Self.Handle,createtime,exittime,kerneltime,usertime) then begin
    OThrdTime:=fThrdTime; fThrdTime:=(int64(kerneltime)+int64(usertime)) div 10000;
    OSysTime :=fSysTime;  fSysTime:=GetTickCount;
    T1:=fThrdTime-OThrdTime;
    T2:=fSysTime-OSysTime;
    if T2<=0 then result:=0 else result:= 100*T1/T2;
  end else result:=-1;
end;

// all of the procedures below can be overridden in descendants
// All of the procedures below are called from this thread, not from VCL !
// Any updates from here to VCL components must be sent via BPM_MsgBufWrite()
// It is strongly advised not to call any thread object functions directly but to
// set parameters using BPM_MsgBufWrite and handling these messages in OnMsgHandler();
procedure TBPM_BaseThread.OnBeforeExecute; begin end; // called when thread is used first
procedure TBPM_BaseThread.OnAfterExecute;  begin end; // called when thread is terminated
procedure TBPM_BaseThread.OnActivate;      begin end; // called when active is set to true
procedure TBPM_BaseThread.OnDeActivate;    begin end; // called when active is set to false
procedure TBPM_BaseThread.OnMaint;         begin end; // called continuously while thread is active
procedure TBPM_BaseThread.OnMaint1000;     begin end; // called every second while thread is active
procedure TBPM_BaseThread.OnMsgHandler;    begin end; // called when thread receives a message
procedure TBPM_BaseThread.OnErrorlog;      begin end; // called when thread code raises an exception

initialization
  EndAllThreadsRequest:=false;
end.
