//______________________________________________________________________________
//
//  Implementatation of a basic messaging  system for decoupling threads or
//  GUI and Engine code
//  Written for an article in Blaise Pascal Magazine
//
//  Feel free to alter and experiment with it.
//  Also feel free to ask questions or contribute or comment to this ideas.
//
//  J.P.Hoefnagel, 1M2M B.V.
//  EMail: jph@1m2m.eu
//______________________________________________________________________________
//

unit uBPM_Messages;

interface uses
  winapi.windows, Winapi.Messages,
  system.syncobjs,system.classes, system.inifiles, system.SysUtils,system.DateUtils,
  System.Variants,
  vcl.forms,vcl.comctrls,vcl.graphics,Vcl.Controls,Vcl.Dialogs,Vcl.StdCtrls,
  Vcl.ExtCtrls,
  registry;

type TBPM_MessageID  =  integer;
type TBPM_CommandID  =  integer;

type  TBPM_MessageRec = record
        MsgID : TBPM_MessageID;
        Cmd   : TBPM_CommandID;
        Msg   : string;
        Color : TColor;
      end;

function  BPM_MsgBufOpen   (aFilter:TBPM_MessageID):THandle;
procedure BPM_MsgBufClose  (aHandle:THandle);
procedure BPM_MsgBufWrite  (aMsgID:TBPM_MessageID; aCmd:TBPM_CommandID;
                            const aMsg: string; aColor: integer);
function  BPM_MsgBufRead   (aHandle:THandle;var aMsgRec:TBPM_MessageRec):boolean;

function  BPM_RegisterMsgID(const MsgIdName:string):TBPM_MessageID;
function  BPM_RegisterCmdID(const CmdIdName:string):TBPM_CommandID;

// Generic command codes
var CmdBtnActiveLed:TBPM_CommandID;

implementation

type TMsgFifo = class(TObject)
     private
       FifoLock  : TCriticalSection;
       Head,Tail : integer;
       buf       : array of TBPM_MessageRec;
     public
       function    Read(var MR:TBPM_MessageRec):boolean;
       procedure   Write(const MR:TBPM_MessageRec);
       constructor Create();
       destructor  Destroy(); override;
     end;

type TBufferRec = record
       MsgFifo   : TMsgFifo;
       Handle    : THandle;
       Filter    : TBPM_MessageID;
     end;
     TBufferArr = array of TBufferRec;

var  Lock        : TCriticalSection;
     slMsg       : TStringlist;
     slCmd       : TStringlist;
     Buffers     : TBufferArr;
     CurrHandle  : integer;

function FindHandleInx(aHandle:THandle):integer;
var n:integer;
begin
  result:=-1; n:=0; while (result<0) and (n<length(Buffers)) do
    if (buffers[n].handle = aHandle) then result:=n else inc(n);
end;

function BPM_MsgBufOpen(aFilter: TBPM_MessageID): THandle;
var i:integer;
begin
  Lock.Enter; try
    inc(CurrHandle);
    i:=length(Buffers); setlength(Buffers,i+1);
    Buffers[i].Handle  := CurrHandle;
    Buffers[i].MsgFIFO := TMsgFifo.Create();
    Buffers[i].Filter  := aFilter;
    result:=CurrHandle;
  finally Lock.leave; end;
end;

procedure BPM_MsgBufClose  (aHandle:THandle);
var n,i,l:integer;
begin
  Lock.Enter; try
    i:=FindHandleInx(aHandle);
    if (i >= 0) then begin // result=true if handle exists in buffers[]
      buffers[i].MsgFifo.Destroy;
      l:=length(Buffers);
      for n:=i to l-2 do buffers[n]:=buffers[n+1];
      setlength(Buffers,l-1);
    end;
  finally Lock.leave; end;
end;

procedure DisposeAll();
var n:integer;
begin
  Lock.Enter; try
    for n:=0 to length(Buffers)-1 do freeandnil(buffers[n].MsgFifo);
  finally Lock.leave; end;
end;

function BPM_MsgBufRead(aHandle:THandle;var aMsgRec:TBPM_MessageRec): boolean;
var i:integer;
begin
  Lock.Enter; try
    i:=FindHandleInx(aHandle);
    result := (i>=0); // result=true if handle exists in buffers[]
    if result then result := (buffers[i].MsgFifo.Read(aMsgRec));
  finally Lock.leave; end;
end;

procedure BPM_MsgBufWriteX(const MR: TBPM_MessageRec);
var i:integer; f:TBPM_MessageID;
begin
  Lock.Enter; try
    for i:=0 to length(buffers)-1 do begin
      f:=buffers[i].Filter;
      if (f = -1) or (f = mr.MsgID) then buffers[i].MsgFifo.Write(mr);
    end;
  finally Lock.leave; end;
end;

procedure BPM_MsgBufWrite(aMsgID:TBPM_MessageID; aCmd:TBPM_CommandID;
                             const aMsg: string; aColor: integer);
var MR:TBPM_MessageRec;
begin
  MR.MsgID :=aMsgID;
  MR.Cmd   :=aCmd;
  MR.Msg   :=aMsg;
  MR.Color :=aColor;
  UniqueString(MR.Msg);
  BPM_MsgBufWriteX(MR);
end;

constructor TMsgFifo.Create;
begin
  FifoLock:=tcriticalsection.Create;
  Head := 0; Tail := 0; setlength(Buf,100);
end;

destructor TMsgFifo.Destroy;
begin
  freeandnil(FifoLock);
  setlength(buf,0);
  inherited;
end;

function TMsgFifo.Read(var MR: TBPM_MessageRec): boolean;
begin
  fifolock.Enter;
  try
    result:=(Head > Tail);
    if result then begin
      MR:=buf[tail];
      inc(Tail);
      if (Tail = Head) then begin
        tail:=0; head:=0;
        if (length(buf)>100) then setlength(Buf,100);
      end;
    end;
  finally
    fifolock.Leave;
  end;
end;

procedure TMsgFifo.Write(const MR: TBPM_MessageRec);
begin
  FifoLock.Enter;
  try
    if Head=length(buf) then begin
      if (head>10000) then begin
        dec(head); // safety, no more than 10000 messages in fifo
      //raise Exception.Create('MdgBuffer fifo Overflow');
      end;
      setlength(buf,Head+100);
    end;
    Buf[Head]:=MR; inc(Head);
  finally
    FifoLock.Leave;
  end;
end;

function BPM_RegisterMsgID(const MsgIdName:string):TBPM_MessageID;
begin
  lock.Enter; try
    result:=slMsg.IndexOf(MsgIDName);
    if (result = -1) then result:=slMsg.Add(MsgIDName);
    Result:=result+$8000;
  finally Lock.Leave; end;
end;

function BPM_RegisterCmdID(const CmdIdName:string):integer;
begin
  lock.Enter; try
    result:=slCmd.IndexOf(CmdIdName);
    if result = -1 then result:=slCmd.Add(CmdIdName);
  Result:=result+$8000;
  finally Lock.Leave; end;
end;

initialization
  CurrHandle      := 1;
  Lock            := TCriticalSection.create();
  slMsg           := tstringlist.create;
  slCmd           := tstringlist.create;
  CmdBtnActiveLed := BPM_RegisterCmdID('BtnActiveLed');
finalization
  DisposeAll();
  FreeAndNil(Lock);
  FreeAndNil(slMsg);
  FreeAndNil(slCmd);
end.





